#!/bin/env python

import argparse
import os
import sys
import textwrap
from maven_artifact.artifact import Artifact

try:
    from maven_artifact.utils import Utils
except ImportError:
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
    from maven_artifact.utils import Utils

from maven_artifact.requestor import RequestException
from maven_artifact.downloader import Downloader


class DescriptionWrappedNewlineFormatter(argparse.HelpFormatter):
    """An argparse formatter that:
    * preserves newlines (like argparse.RawDescriptionHelpFormatter),
    * removes leading indent (great for multiline strings),
    * and applies reasonable text wrapping.

    Source: https://stackoverflow.com/a/64102901/79125
    """

    def _fill_text(self, text, width, indent):
        # Strip the indent from the original python definition that plagues most of us.
        text = textwrap.dedent(text)
        text = textwrap.indent(text, indent)  # Apply any requested indent.
        text = text.splitlines()  # Make a list of lines
        text = [textwrap.fill(line, width) for line in text]  # Wrap each line
        text = "\n".join(text)  # Join the lines again
        return text


class WrappedNewlineFormatter(DescriptionWrappedNewlineFormatter):
    """An argparse formatter that:
    * preserves newlines (like argparse.RawTextHelpFormatter),
    * removes leading indent and applies reasonable text wrapping (like DescriptionWrappedNewlineFormatter),
    * applies to all help text (description, arguments, epilogue).
    """

    def _split_lines(self, text, width):
        # Allow multiline strings to have common leading indentation.
        text = textwrap.dedent(text)
        text = text.splitlines()
        lines = []
        for line in text:
            wrapped_lines = textwrap.fill(line, width).splitlines()
            lines.extend(subline for subline in wrapped_lines)
            if line:
                lines.append("")  # Preserve line breaks.
        return lines


class MainCommand:
    def _get_arguments(self):
        parser = argparse.ArgumentParser(formatter_class=WrappedNewlineFormatter, epilog=__epilog__)
        parser.add_argument(
            "maven_coordinate",
            help="""
            defined by http://maven.apache.org/pom.html#Maven_Coordinates. The possible options are:
            - groupId:artifactId:version
            - groupId:artifactId:packaging:version
            - groupId:artifactId:packaging:classifier:version""",
        )
        parser.add_argument(
            "filename",
            nargs="?",
            help="""
            If not supplied the filename will be <artifactId>.<extension>.
            The filename directory must exist prior to download.""",
        )
        parser.add_argument(
            "-m",
            "--maven-repo",
            dest="base",
            default="https://repo.maven.apache.org/maven2/",
            help="Maven repository URL (default: https://repo.maven.apache.org/maven2/)",
        )

        parser.add_argument("-u", "--username", help="username (must be combined with --password)")
        parser.add_argument(
            "-p",
            "--password",
            help="""
            password (must be combined with --username) or
            base64 encoded username and password (can not not be combined with --username)""",
        )
        parser.add_argument(
            "-t", "--token", help="OAuth bearer token (can not be combined with --username or --password)"
        )

        parser.add_argument("-ht", "--hash-type", default="md5", help="hash type (default: md5)")

        args = parser.parse_args()

        username = args.username
        password = args.password
        token = args.token

        if username and not password:
            parser.error("The 'username' parameter requires the 'password' parameter.")
        elif (username or password) and token:
            parser.error("The 'token' parameter cannot be used together with 'username' or 'password'.")
        elif (password) and not (username or token) and not Utils.is_base64(password):
            parser.error("The 'password' parameter must be base64 if not used together with 'username'.")

        return args


__epilog__ = """
            Example:
                %(prog)s "org.apache.solr:solr:war:3.5.0"\n
            """


def main():
    mc = MainCommand()
    args = mc._get_arguments()

    try:
        dl = Downloader(base=args.base, username=args.username, password=args.password, token=args.token)

        artifact = Artifact.parse(args.maven_coordinate)

        filename = args.filename

        if dl.download(artifact, filename, args.hash_type):
            sys.exit(0)
        else:
            print("Download failed.")
            sys.exit(1)
    except RequestException as e:
        print(e.msg)
        sys.exit(1)


if __name__ == "__main__":
    main()



import base64
import requests

from maven_artifact.utils import Utils


class RequestException(Exception):
    def __init__(self, msg):
        self.msg = msg


class Requestor(object):
    def __init__(self, username=None, password=None, token=None, user_agent="Maven Artifact Downloader/1.0"):
        self.user_agent = user_agent
        self.username = username
        self.password = password
        self.token = token

    def request(self, url, onFail, onSuccess=None, method: str = "get", **kwargs):
        headers = {"User-Agent": self.user_agent}

        if self.username and self.password:
            token = self.username + ":" + self.password
            headers["Authorization"] = f"Basic {base64.b64encode(token.encode()).decode()}"
        elif Utils.is_base64(self.password):
            headers["Authorization"] = f"Basic {self.password}"
        elif self.token:
            headers["Authorization"] = f"Bearer {base64.b64encode(self.token.encode()).decode()}"

        try:
            response = getattr(requests, method)(url, headers=headers, **kwargs)
            response.raise_for_status()
            if onSuccess:
                return onSuccess(response)
            return response
        except Exception as ex:
            if onFail:
                return onFail(url, ex)
            raise



import pytest
from maven_artifact import Artifact


def test_artifact_missing_params():
    with pytest.raises(TypeError):
        Artifact()


def test_issnapshot():
    artifact = Artifact(
        group_id="com.example", artifact_id="example-artifact", version="0.0.1", classifier="javadoc", extension="zip"
    )

    assert artifact.is_snapshot() is False

    artifact = artifact.with_version(_version="0.0.1-SNAPSHOT")

    assert artifact.is_snapshot() is True



import os
import tempfile

from maven_artifact import Downloader
from maven_artifact.artifact import Artifact


def test_downloader_of_existing_artifact():
    artifact = Artifact.parse("org.apache.solr:solr:war:3.5.0")

    dl = Downloader()

    tmpdirname = tempfile.TemporaryDirectory()

    tmpfile = os.path.join(tmpdirname.name, "example.war")

    dl.download(artifact, filename=tmpfile)

    assert os.path.exists(tmpfile)

    tmpdirname.cleanup()



from lxml import etree

from maven_artifact.requestor import RequestException


class Resolver(object):
    def __init__(self, base, requestor):
        self.requestor = requestor
        if base.endswith("/"):
            base = base.rstrip("/")
        self.base = base

    def _find_latest_version_available(self, artifact):
        path = "/%s/maven-metadata.xml" % (artifact.path(False))
        xml = self.requestor.request(self.base + path, self._onFail, lambda r: etree.fromstring(r.content))
        v = xml.xpath("/metadata/versioning/versions/version[last()]/text()")
        if v:
            return v[0]

    def _find_latest_snapshot_version(self, artifact):
        path = "/%s/maven-metadata.xml" % (artifact.path())
        xml = self.requestor.request(self.base + path, self._onFail, lambda r: etree.fromstring(r.content))
        timestamp = xml.xpath("/metadata/versioning/snapshot/timestamp/text()")[0]
        buildNumber = xml.xpath("/metadata/versioning/snapshot/buildNumber/text()")[0]
        meta_version = f"{timestamp}-{buildNumber}"
        # obtain exact version value
        versions = xml.xpath("/metadata/versioning/snapshotVersions/snapshotVersion/value/text()")
        for version in set(versions):
            if meta_version in version:
                return version
        return meta_version

    def _onFail(self, url, e):
        raise RequestException(f"Failed to download maven-metadata.xml from {url} due to {e}")

    def resolve(self, artifact):
        version = artifact.version
        if not artifact.version or artifact.version == "latest":
            version = self._find_latest_version_available(artifact)
        elif artifact.is_snapshot():
            version = self._find_latest_snapshot_version(artifact)
        return artifact.with_version(version)

    def uri_for_artifact(self, artifact):
        resolved = self.resolve(artifact)
        return artifact.uri(self.base, resolved.version)



import hashlib
import os

from maven_artifact.requestor import RequestException, Requestor
from maven_artifact.resolver import Resolver


class Downloader(object):
    def __init__(self, base="https://repo.maven.apache.org/maven2", username=None, password=None, token=None):
        self.requestor = Requestor(username=username, password=password, token=token)
        self.resolver = Resolver(base, self.requestor)

    def download(self, artifact, filename=None, hash_type="md5"):
        filename = artifact.get_filename(filename)
        url = self.resolver.uri_for_artifact(artifact)
        if os.path.exists(filename) and self.verify_file(filename, url, hash_type=hash_type):
            print("%s is already up to date" % artifact)
            return artifact

        print(f"Downloading artifact {artifact} from {url}")

        def onError(uri, err):
            self._throwDownloadFailed(f"Failed to download {artifact} from {uri} \n{err}")

        with self.requestor.request(url, onError, stream=True) as r:
            with open(filename, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        print(f"Maven artifact {artifact} is downloaded to {filename}")
        return artifact

    def _throwDownloadFailed(self, msg):
        raise RequestException(msg)

    def verify_file(self, file, url, hash_type: str = "md5"):
        url = f"{url}.{hash_type}"

        def onError(uri, err):
            self._throwDownloadFailed("Failed to download hash file from " + uri)

        remote_hash = self.requestor.request(url, onError, lambda r: r.text)
        local_hash = getattr(hashlib, hash_type)(open(file, "rb").read()).hexdigest()
        return remote_hash == local_hash



import os


class Artifact(object):
    def __init__(self, group_id, artifact_id, version, classifier=None, extension=None):
        if not group_id:
            raise ValueError("group_id must be set")
        if not artifact_id:
            raise ValueError("artifact_id must be set")

        self.group_id = group_id
        self.artifact_id = artifact_id
        self.version = version
        self.classifier = classifier
        self.extension = extension or "jar"

    def is_snapshot(self):
        return self.version and self.version.endswith("-SNAPSHOT")

    def path(self, with_version=True):
        base = self.group_id.replace(".", "/") + "/" + self.artifact_id
        return base if not with_version else f"{base}/{self.version}"

    def uri(self, base, resolved_version=None):
        if self.is_snapshot() and not resolved_version:
            raise ValueError("Expected unique version for snapshot artifact " + str(self))
        elif not self.is_snapshot():
            resolved_version = self.version
        ret = f"{base}/{self.path()}/{self.artifact_id}-{resolved_version}"
        if self.classifier:
            ret += "-" + self.classifier
        return f"{ret}.{self.extension}"

    def with_version(self, _version):
        return Artifact(self.group_id, self.artifact_id, _version, self.classifier, self.extension)

    def _generate_filename(self):
        if not self.classifier:
            return f"{self.artifact_id}.{self.extension}"
        return f"{self.artifact_id}-{self.classifier}.{self.extension}"

    def get_filename(self, filename=None):
        if not filename:
            filename = self._generate_filename()
        elif os.path.isdir(filename):
            filename = os.path.join(filename, self._generate_filename())
        return filename

    def __str__(self):
        if self.classifier:
            return "%s:%s:%s:%s:%s" % (self.group_id, self.artifact_id, self.extension, self.classifier, self.version)
        elif self.extension != "jar":
            return "%s:%s:%s:%s" % (self.group_id, self.artifact_id, self.extension, self.version)
        else:
            return "%s:%s:%s" % (self.group_id, self.artifact_id, self.version)

    @staticmethod
    def parse(maven_coordinate):
        parts = maven_coordinate.split(":")
        if len(parts) >= 3:
            g = parts[0]
            a = parts[1]
            v = parts[len(parts) - 1]
            t = None
            c = None
            if len(parts) == 4:
                t = parts[2]
            if len(parts) == 5:
                t = parts[2]
                c = parts[3]
            return Artifact(group_id=g, artifact_id=a, version=v, classifier=c, extension=t)
        else:
            return None



from maven_artifact.requestor import Requestor, RequestException  # noqa: F401
from maven_artifact.artifact import Artifact  # noqa: F401
from maven_artifact.resolver import Resolver  # noqa: F401
from maven_artifact.downloader import Downloader  # noqa: F401
from maven_artifact.utils import Utils  # noqa: F401



import base64


class Utils:
    @staticmethod
    def is_base64(sb):
        try:
            if isinstance(sb, str):
                # If there's any unicode here, an exception will be thrown and the function will return false
                sb_bytes = bytes(sb, "ascii")
            elif isinstance(sb, bytes):
                sb_bytes = sb
            else:
                raise ValueError("Argument must be string or bytes")
            return base64.b64encode(base64.b64decode(sb_bytes)) == sb_bytes
        except Exception:
            return False
